/** \file main.c
 *
 * \brief This program implements a protocol for a command line interface.
 *
 * \par Protocol Definition
 *
 * * Start Condition: Command (C, N)
 * * Stop Condition: Terminal (*space*)
 * * Message Format: 'COMMAND''ARGUMENT''TERMINAL' or 'COMMAND''ARGUMENT1''ARGUMENT2''TERMINAL'
 *   * C *char* - Read a character
 *   * N *digit* - Read a number
 *   * N *digit* *digit* - Read a number
 *
 * \par
 * The finite state machine for this protocol can be found in the enclosed .PNG-file. The program
 * loops until CTRL-C is pressed.
 *
 * \author    Wolfgang Neff
 * \version   1.0
 * \date      2018-11-24
 *
 * \par History
 *      Created: 2018-11-24
 */

#include <ctype.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>

#define BREAK '\x03'

/// <summary>Command enumeration.</summary>
/// <remarks>
/// This <c>typedef</c> defines an enumeration data type for the
/// available commands. The elements of the enumeration can be
/// as constants.
/// </remarks>
typedef enum {
	CMD_NUMBER,
	CMD_CHARACTER
} command_t;

/// <summary>State enumeration.</summary>
/// <remarks>
/// This <c>typedef</c> defines an enumeration data type for the
/// different states. The elements of the enumeration can be
/// as constants.
/// </remarks>
typedef enum {
	STATE_GET_COMMAND,
	STATE_GET_CHAR_ARG,
	STATE_GET_NUM_ARG1,
	STATE_GET_NUM_ARG2,
	STATE_GET_TERMINAL,
	STATE_EXECUTE,
	STATE_ERROR
} state_t;

/// <summary>Numeric argument</summary>
/// <remarks>
/// This variable is used to store the numeric argument.
/// </remarks>
int num_arg;

/// <summary>Character arguments</summary>
/// <remarks>
/// This variable is used to store the character argument.
/// </remarks>
char char_arg;

/// <summary>Command</summary>
/// <remarks>
/// This variable is used to store the command.
/// </remarks>
command_t command;

/// <summary>State</summary>
/// <remarks>
/// This variable is used to store the current state.
/// The initial state is <c>STATE_GET_COMMAND</c>
/// </remarks>
state_t state = STATE_GET_COMMAND;

/// <summary>Get next token.</summary>
/// <remarks>
/// A token is a character with an assigned meaning. This routine
/// reads the next character from the console and returns it. All
/// characters are supposed to have a meaning. Function <c>_getche()</c>
/// from <c>conio.h</c> is used to read a key. The program will
/// exit if the BREAK key (CTRL-C) is pressed.
/// </remarks>
/// <returns>The next token.</returns>
char get_token(void)
{
	char token;
	token = _getche();
	if (token == BREAK) exit(EXIT_SUCCESS);
	return token;
}

int main(void)
{
	char token;
	while (1) {
		switch (state) {
		case STATE_GET_COMMAND:
			// In state STATE_GET_COMMAND we check if a valid token has been provided.
			// If a valid command has been found it will be stored. The subsequent
			// state depends on the command found.
			token = toupper(get_token());
			if (token == 'C') {
				// The command is: read a character.
				command = CMD_CHARACTER;
				// In the subsequent state we have to read a character.
				state = STATE_GET_CHAR_ARG;
			}
			else if (token == 'N') {
				// The command is: read a number.
				command = CMD_NUMBER;
				// In the subsequent state we have to read the first digit of the number.
				state = STATE_GET_NUM_ARG1;
			}
			else {
				// The specified token is no valid command.
				state = STATE_ERROR;
			}
			break;
		case STATE_GET_CHAR_ARG:
			token = get_token();
			if (isalpha(token)) {
				// We only accept alphabetic tokens. The character will be stored
				char_arg = token;
				// There is only one argument. We are finished. We are looking
				// for the terminal symbol.
				state = STATE_GET_TERMINAL;
			}
			else {
				state = STATE_ERROR;
			}
			break;
		case STATE_GET_NUM_ARG1:
			token = get_token();
			if (isdigit(token)) {
				// We only accept decimal tokens. The token is a character.
				// We have to subtract '0' in order to convert it into a
				// number.
				num_arg = token-'0';
				// There may be another digit. We are not yet finished. We
				// are looking for another argument.
				state = STATE_GET_NUM_ARG2;
			}
			else {
				state = STATE_ERROR;
			}
			break;
		case STATE_GET_NUM_ARG2:
			// The second argument is optional. We accept either a decimal
			// or a terminal.
			token = get_token();
			if (isdigit(token)) {
				// We found a decimal. The argument consists of two digits.
				// We calculate its value by 10*arg1 + arg2. arg1 ist stored
				// in num_arg, arg2 is stored as ASCII value in token.
				num_arg = 10*num_arg + (token - '0');
				// We found two digits. We are finished. We are looking
				// for the terminal symbol, now.
				state = STATE_GET_TERMINAL;
			}
			else if (isblank(token)) {
				// We found a terminal. The argument consists of only one digit.
				// The terminal has yet been found. Next state is STATE_EXECUTE.
				state = STATE_EXECUTE;
			}
			else {
				state = STATE_ERROR;
			}
			break;
		case STATE_GET_TERMINAL:
			// We expect the terminal to be a blank character.
			token = get_token();
			if (isblank(token)) {
				state = STATE_EXECUTE;
			}
			else {
				state = STATE_ERROR;
			}
			break;
		case STATE_EXECUTE:
			// A terminal character has been found. Now we are ready to
			// execute the command with its argument.
			if (command == CMD_CHARACTER) {
				printf("\n");
				printf("Command: Get Character, Character: %c", char_arg);
				printf("\n");
			}
			else if (command == CMD_NUMBER) {
				printf("\n");
				printf("Command: Get Number, Value %d", num_arg);
				printf("\n");
			}
			// We are ready for the next command.
			state = STATE_GET_COMMAND;
			break;
		case STATE_ERROR:
			// An error has occured. We report it and resume.
			printf("\n");
			printf("Parsing Error!");
			printf("\n");
			state = STATE_GET_COMMAND;
			break;
		}
	}
}
